// ppheader.hpp  General header file for PP                  11aug98-26jly01
//                   SQUASH-POP WORLD       Landmarks
// NOTE. short and long, not int, are used as they stay the same for
//       16bit and 32bit data types. BC4PG 21. Exceptions are where a
//       library function requires int.

//  Stimulus-predicting associations have last-node-lists instead of
//  back-pointers and pointer-probs.     4oct99
//  Last-node-lists are limited to max_last_node_list and nodes are
//  added from the front.   9oct99
//  BodyMove and EyeMove actions merged into one to save memory nets. Feel
//   dropped and contexts become totally visual. \bodyeye   23oct99
//  Actions put on choice lists only if they point to a node with higher
//   expectation than that for the last step.  \uphill     2nov99
//  Tried this but it gave much worse results so it was abandoned:
//     [Counts as uphill if points to node with max expn (expn=age)]  17nov99
//  Found that novelty was being zeroed instead of decremented by plans,
//     which may account for absence of nov=3 and nov=2 associations. 19nov99
//     Corrected this and restarted. Since this might have caused the strange
//     failure of counting max expn as uphill, I restored that as well, but
//     will cancel it again if it doesn't improve things.
//  Changed context_links so that they have a separate pointer to the
//     association instead of casting a pointer to a context_link. 26nov99
//  The idea is to limit the number of bptrs, not worrying about the fact that
//     the network will not be entirely consistent. Latest ptr is always stored
//     at the front of the list and if different from existing ones it can push
//     out the one at the end. 28nov99
//  Not happy about performance of fovea predicting templates. FW is being
//     dropped to increase speed and because it does not seem to be very
//     effective. The ambiguity due to the movement of the robots will always
//     be there in such a small world.Trying to strengthen some of the other
//     fovea predicting templates. Leaving out last-node pointers, too,
//     as they don't seem to be doing anything either! 29nov99
//  In calculating worths, nov>1 has value 1 and nov=1 has value expectation-
//     for-novelty.   30nov99
//  Found that Boredom wasn't being incremented when expn dropped, which was
//     causing the runs to go into loops. 6dec99
//  Found that in T&T canvas cell 4 was being updated by Fovea. Restart! 8dec99
//  Suspect that crashing after about step 70000 is due to leakback active, fixed
//     and holding-lists demanding too much memory. There could be an overhead
//     of 20Bytes on each item "new"ed. Trying to reprogram these lists as fixed
//     arrays which are not allocated dynamically, but just filled and zeroed.
//     Also memory will be saved by having a single assoc ptr in place of a
//     link.  14dec99.
//  Separated memory specific to robot A (pp_info *pp) from the memories
//     (robot_info *sa and robot_info *sb) which are the same for both robots,
//     so that it is not duplicated and not swapped. 16dec99
//  Predictions in two (stimulus and action) arrays. 30dec99
//  Transition probabilities in two (stimulus and action) arrays. 01jan00
//  Backpointers in an array. 02jan00
//  One cyclic list used for leakback (PMA suggestion) instead of 3 lists. Also,
//     instead of levels, have a limitation on number of nodes processed.3jan00
//  Array used for context links. 04jan00
//  System with canvas in BE template showed no signs of reaching a limit of
//     no of associations, so this version does without canvas. The first run
//     couldn't hold reward marks, so reward_novelty was increased
//     from 5 to 20.  11jan00
//  On 4feb00 repeated 25jan00 run but with leakback using bkptrs to same node
//     to give more correct probabilities.
//  On 8feb00 ran program without learning to get average reward rate of 132.2
//  On 9feb00 ran program with templateS BE,W1 having FOVNODIRN replaced by
//     FOVNOEYE to see if this improved PP's ability to predict movements of
//     robot B, as this will give the direction of a robot under A's fovea.
//     No print-out except every 10000 steps with dumping of memory.
//     Backpointers to same node used in leakback.
//  On 12feb00  Run with canvas replaced by touch. Reward rate almost to
//     500 per 10k steps, but few plans probably because of inadequate
//     touch-predicting contexts.
//  On 16feb00  Enlarged context template for predicting touch.
//  Also 16feb00 New canvwide directory and changing canvas stimuli to
//     support wide predictions for each direction of eye move.
//  24fb00 Noticed that expectations are too high throughout ClBE. Corrected
//     (I think) error where predptr->novelty was used as a multiplying factor
//     even though it could be as high as 20.
//     Introduced w.expn_calc_omits_ptrs_to_same_node flag which prevents use of
//     bkptr to same node as well as forward ptr to same node in leakback calc.
//  28feb00 Noticed that subsumption eye moves were not according to algorithm
//     and found that part of program was not taking into account body move.
//     re-run with correction.
//  01mar00 The small number of plans to reward in the 28feb00 run, in spite of
//     the high rate of reward, suggested that plans were failing because only
//     the most probable predictions of stimuli were being used and these were
//     often incompatible with realizable template BE contexts. Therefore, when
//     there is no match with the BE template in T+T, alternative stimulus
//     predictions should be tried. Choose in order of remaining highest prob.
//     of prediction. This was done in the run of 7mar00 with dramatic results.
//  08mar00 Plan marks in memory should be numbered so that a match is recorded
//     only if an event matches a mark for the right step in memory.
//  12,14mar00 - major runs with numbered marks for plans, and use of
//     alternative stimuli in plans. Stored in marknums\xx folder.
//  19mar00 - same program and parameters, but with no learning.
//  19mar00 (later) - same as 12,14mar00 with different seed.
//  20mar00 FOVNODIRN replaced by FOVEA in context of template BE.
//  25mar00 Same as 20mar00 with different random seed.
//  28mar00 Corrected mistake in implementation of help-A strategy.
//  01apr00 Runs with no help-A strategy. First no-learning, then with learning.
//  10apr00 Increased resolution of Proprioception and included it in BE
//          context. Robot B helping.  tt\*.dat
//  17apr00 Strengthened B-helps-A strategy using propriocep when fovea on O.
//  18apr00 Probability of Reward by reward counters as well as novelty in
//          predictions to distinguish associations that are rewarded
//          consistently from those that are not.
//  20apr00 Increased resolution of propriocep to 5x5 cells inside NN,EE,SS,WW.
//  26apr00 Same with no learning.
//  27apr00 Same as 20apr00 with leakback limit set to 3000 nodes.
//  03may00 New leakback starting with all nov>0 nodes on active list and
//          working in levels, although still using single list. Also does
//          so many levels per step to minimize Windows bomb-out.
//  11may00 Speeded up leakback using an array.
//  16may00 Improving nov=20 decrementation process.
//  20may00 Decreasing significance of past memory by increasing quantum.
//  11jun00 includes correction of abs() error and leaves quantum constant.
//  18jun00 (a) No Learning in ww\ww*.dat
//          (b) Subsump actns, no plans, else same for 0-100k steps in xx\xx*.dat.
//          (c) Subsump actns, no plans, no reward for 0-100k steps in yy\yy*.dat
//          (d) No rewards for 0-100k steps in zz\zz*.dat
//  19jun00 Parameters for controlling 0-N steps. Printing out for Matlab m-file.
//  02jly00 Return to Windows after NODES_BETWEEN_WINDOWS nodes of leakback.
//  02jly00 Return to Windows after NODES_BETWEEN_WINDOWS nodes of leakback.
//  14jly00 Food and drink leads to confused learning.
//  17jly00 Trying food only.
//  19jly00 Food only seems to be OK so going back to food and drink on
//  20jly00 with hunger reducing resource every F action, thirst every action.
//  24jly00 Trying to get a balance between hunger and thirst. First successful
//          run. Is in foodrink\both\vv*.dat and runs to 2010000 steps.
//  25jly00 uu*.dat is the same as vv*.dat, omitting some of the detailed
//          sequences but going on to 10m steps?
//  28jly00 Same again with a slight reduction in resource inc/reward: tt*.dat
//  31jly00 Same again as 28jly00 with eye-move algorithm giving preference to
//            moving to wide for obj/robot over staying on fovea for obj/rob.
//  01aug00 Back to staying on fovea before moving to wide for object, then rob.
//          random(10) to random(20) in ppworld.cpp to make robot B less random.
//  05aug00 Put FOVNODIRN into W0ClW template. Otherwise same as 01aug00.
//  07aug00 Removed direction and eye from fovea and ran.  dirnstim\zz*.dat
//  09aug00 Direction of robot B (if in fovea or wide) as new stimulus. \yy*.dat
//  10aug00 Simplified robot B's helpful moving strategy. dirnstim\xx*.dat
//  11aug00 Reduced Fovea and Wide predicting templates to one each. No canvas
//            cells used and both Fovea and Wide in contexts to help predict
//            O and Q more consistently.  dirnstim\ww*.dat
//  15feb01 DOS window and forgetting programs from SideBy Win32 programs.
//  16feb01 First successful run. sp16fb01.m   zz*.dat
//  17feb01 Counting to 10k steps  yy*.dat
//  18feb01 Restricting memory xx*.dat ctxtsA 2200, ctxtsS 3300, predsA&S 6000
//            ptrprbs 9000, bptrs 9000, linksA&S 7000
//  22feb01 Restricting ctxts only ww*.dat ctxtsA 2000, ctxtsS 300, predsA&S 7000
//            ptrprbs 10000, bptrs 10000, linksA&S 8000
//  27,28feb01 TellMove with BE context Wide,Fovea,Direction,Touch,Hear and
//            SA context Wide,Fovea,Direction,Touch. d:\ppdata\tellmove\zz27fb01.zip
//  28feb01 Dropped Touch from SA context to allow PP to learn alternative speech
//          actions. Also allowing PP to say even when no drive. tellmove\yy*.dat
//  05mar01 HEAR, HEARSF, HEAROR, HEARFO stimuli to enable PP to distinguish sounds
//          it has made itself from outside sounds.
//  07mar01 Everything being forgotten more than 4000 steps back, so increased
//          limits on nodes from A2000,S3000 to A5000,S5000
//  08mar01 Proprioception added to contexts of Touch and Wide. Unable to get
//          of subsumption level.
//  09mar01 Proprioception removed again from contexts of Touch and Wide to
//          check that nothing else keeping it from rising above subsumption
//          level. OK, so changed eye algorithm to prefer moving to wide over
//          staying on fovea when object or robot there.
//  12mar01 Landmark template does not take part in plan making.
//  13mar01 Changed resource decrements to proportional to make them nonlinear
//          and make the system less likely to drop below subsumption-only level.
//  17mar01 Playing around with Desire and Landmarks.
//  26mar01 Routine in update_nodeS to remove preds without current landmark.
//  02apr01 Dumping etc determined by step series.
//  09apr01 Landmarks adjusted only if a drive is on.
//  10apr01 Nov=2,4,6 not now a criterion for a landmark.  xxall.zip
//  12apr01 Inc/dec of landmark-tAge reduced to 1 from quantum,
//          and mark is 2,4,6   ww*.dat
//  17apr01 Increasing reward-rate-threshold but only 50% chance of
//          novelty-removal if below threshold. Also reduced MAX_NODES_A to 5000
//          as only 6000 nodes used up to about 3.5m steps. vv*.dat
//  24apr01 Moved from examples\landmark to examples\foodrink to try
//          moving update_ltmA to after get_stimuli_HN so that "GOOD"
//          reward can affect storing. Should also simplify mimic-speech.
//  25apr01 Robot B is telling A to make moves when not hungry/thirsty.
//          Remember that I am letting robot B see if A is hungry/thirsty!!<<<<
//  04may01 w.ul_B_told_A not nulled properly in xx1may01 run. Corrected in zz.
//  07may01 Found a mistake in DN template and corrected this. Changed DE. yy*.dat
//  08may01 Computer had turned itself off several times. Switched right off,
//          defragmented and restarted.  xx*.dat
//  09may01 Error again probably due to screensaver. Turned off screensaver and
//          power-saving screen-turn-off. Will do it manually. Added Desire event
//          type to W, F and DN templates, BodyMove to DE. zz*.dat again
//  12may01 9may01 run had trouble predicting desire so changed it to timing
//          Also robot B tells A to turn to enable B to squash object. yy*.dat
//  15may01 zz*.dat
//  18may01 Removed Touch replaces Desire in F,W,DN temp. No landmarks. yy*.dat
//  24may01 Found that shifting ppltmA invalidated mark checking. zz*.dat
//  29may01 More corrections to get mark-checking and plans to GOOD right yy*.dat
//  01jun01 Wipe_all_marks not doing BE and SA separately. Corrected.
//  06jun01 Marks must be set to NULLMARK (= 0).  yy*.dat
//  07jun01 Good run with GOOD reward for single action copy. yy*.dat
//  10jun01 Reward for 2 consecutive action copies by robot A. zz*.dat
//  22jun01 Changed program so that extra graph line can be obtained.
//  25jun01 ClSS associations had been unintentionally restricted in 10jun01 run,
//       so node limit increased from 50 to 200, pred limit from 100 to 1000.
//  13jly01 HEAR added to SA and replaces HEARFO in BE template   xx*.dat
//       This is not as good a run as 25jun01 in spite of better plans at step
//       500k. My guess is that BE and SA clusters need more associations.
//  16jly01 Upper template to predict DIRECTION. HEARFO back in BE, HEAR out of
//       SA. ww*.dat
//  26jly01 Corrected marking proc in Ppreda.cpp, otherwise same 16jly01 vv*.dat


#ifndef	_PPHEADER_H
	#define	_PPHEADER_H
#include <stdio.h>
#include <conio.h>
#include <iostream.h>
#include <fstream.h>      // See BC4 Programmer's Guide op.214,5
#include <string.h>
#include <cstring.h>
#include <stdlib.h>
#include <time.h>

#define TRUE			 1    // Not CG's -1, to be compatible with BC4PG103
#define FALSE            0
#define H_N              0                   // Here and Now
#define T_T              1                   // Theree and Then
#define AC               1                   // Action
#define ST               2                   // Stimulus
#define BOTH             3                   // Action and Stimulus
#define TIM              1                   // Timing
#define THR              0                   // Threading
#define NIL              0
#define CHOICE			 1
#define RECENCY			 2
#define FOOD             2            // for food reward
#define DRINK            4            // for drink reward
#define PLUS             6            // for food plus drink
#define GOOD             8            // for "GOOD" speech reward
#define AGEWORTHONLY	 	 0            // For choosing stimuli  25sep99
#define EXPNWORTHONLY	 1            // For choosing stimuli  25sep99
#define WORTHANDAGEWORTH 2            // For choosing actions  25sep99
#define HANDFEED		  200
#define QUANTUM                        10         //Initial value 220may00
#define SCALE_FACTOR				   10000		  // 13may99
#define DIVSF						   1000       // Divn scale factor
#define MAX_INPUT_STRING	           10000		  // 20may99
#define MAX_EVENT_TYPE_NAME            12
#define TEMPLATE_NAME_LENGTH           5
#define CLUSTER_NAME_LENGTH            5
#define MAX_NODES_A                    5000     // 300
#define MAX_NODES_S                    5000     // 300
#define MAX_PREDICTIONS_A              16000    // 400
#define MAX_PREDICTIONS_S              16000    // 400
#define MAX_POINTER_PROBS              32000L   // 500
#define MAX_BACK_POINTERS              32000L   // 500
#define MAX_CONTEXT_LINKS_A            20000   // 1000
#define MAX_CONTEXT_LINKS_S            15000   // 1000
#define MAX_TEMPLATES                  11
#define MAX_A_CLUSTERS                 2
#define MAX_S_CLUSTERS                 8
#define MAX_CLUSTERS           MAX_A_CLUSTERS + MAX_S_CLUSTERS
#define MAX_LEAKBACK_CLUSTERS          2
#define MAX_TEMPS_IN_CLUSTER           1         // MAX TEMPLATES IN CLUST
#define MAX_CONTEXT_EVENTS             6
#define MAX_EVENT_SHIFTS               3
#define MAX_EVENTS_IN_STREAM           1
#define MAX_TEMP_PREDICTION_TYPES      2
#define MAX_EVENT_TYPES                30
#define MAX_A_STREAMS                  2          //  2 with speech 16sep99
#define MAX_S_STREAMS                  8          //  6 with speech 16sep99
#define MAX_STREAMS             MAX_A_STREAMS+MAX_S_STREAMS
#define MAX_NODES_PER_CHOICE           5          // 16dec99
#define MAX_CHOICES                   10          // 16dec99
#define MAX_REWARDED_PLANS            5000     // 500
#define MAX_REWARD_NOVELTY            15    // 2 for food,4 for drink,8 for good
#define MAX_RANDOM_NODES              20
#define MAX_PLAN_LENGTH               20
#define MAX_MARKS_A             MAX_PLAN_LENGTH*MAX_NODES_PER_CHOICE  // 25jly01
#define MAXDUMMY                      100         // memory separator
#define LKBK_LISTS_LENGTH             15000   // 1500
#define MAX_FILE_LINES                80
#define MAX_REWARD_GETTING_ASSOCIATIONS    15000 // 1200 // per graph dump
#define FOVEA_CELL					  12
#define NORTH_CELL					  7
#define EAST_CELL					  13
#define SOUTH_CELL					  17
#define WEST_CELL					  11
#define NW_CELL						  6
#define NE_CELL						  8
#define SE_CELL						  18
#define SW_CELL						  16
#define NN_CELL						  2
#define EE_CELL						  14
#define SS_CELL						  22
#define WW_CELL						  10


#define HALT            "HALT"       // USED IN ask_auto()
#define ON              1
#define OFF             0
#define NEW_CLUSTER     1            // Used to control return
#define NEW_LEVEL       2            // from leakback to Windows
#define NEW_NODE        3            // and back
#define NOVEL           1            // T&T TERMINATOR
#define NULLWORTH			-32000		 // instead of 0 which is a valid worth.
#define NULLTYPE			-1			    // instead of 0 which is a valid type.
#define NULLEVENT       0XFFFFFFFF	 // must be an invalid event <<<!!!***
#define NULLNODE        -1           // because arrays start at 0
#define NULLLONG        -1L          // because arrays start at 0
#define NULLLINK        -1L          // because arrays start at 0
#define NULLOBJECT      -1           // because object names are all positive.
#define NULLSTREAM      -1           // because arrays start at 0
#define NULLSTEP        -1           // to mark contexts on empty list
#define NULLMARK        -50          // must be less than pp->path_step - 1 (25jly01)
#define LOOPMARK        100          // indicates that T&T is in a loop
#define DONTKNOW        0X00000063	 // must be a unique event for canvas:" ? "

#define BOARD_ROWS		3			// excluding 2 wall rows
#define BOARD_COLS		3 			// excluding 2 wall columns
#define BOARD_CELLS		25			// including wall rows and columns

#define ULONG_0			0X00000030		// unsigned long ASCII equivalents
#define ULONG_1			0X00000031		// unsigned long ASCII equivalents
#define ULONG_2			0X00000032
#define ULONG_3			0X00000033
#define ULONG_4			0X00000034
#define ULONG_5			0X00000035
#define ULONG_6			0X00000036
#define ULONG_7			0X00000037
#define ULONG_8			0X00000038
#define ULONG_9			0X00000039
#define ULONG_a			0X00000061
#define ULONG_b			0X00000062
#define ULONG_A			0X00000041
#define ULONG_B			0X00000042
#define ULONG_C			0X00000043
#define ULONG_D			0X00000044
#define ULONG_E			0X00000045
#define ULONG_F			0X00000046
#define ULONG_G			0X00000047
#define ULONG_H			0X00000048
#define ULONG_I			0X00000049
#define ULONG_J			0X0000004A
#define ULONG_K			0X0000004B
#define ULONG_L			0X0000004C
#define ULONG_M			0X0000004D
#define ULONG_N			0X0000004E
#define ULONG_O			0X0000004F
#define ULONG_P			0X00000050
#define ULONG_Q			0X00000051
#define ULONG_R			0X00000052
#define ULONG_S			0X00000053
#define ULONG_T			0X00000054
#define ULONG_U			0X00000055
#define ULONG_V			0X00000056
#define ULONG_W			0X00000057
#define ULONG_X			0X00000058
#define ULONG_x		    0X00000078
#define ULONG_Y			0X00000059
#define ULONG_y         0X00000079
#define ULONG_Z			0X0000005A
#define ULONG_z			0X0000007A
#define ULONG_AA			0X00004141
#define ULONG_BB			0X00004242
#define ULONG_EE			0X00004545
#define ULONG_FF			0X00004646
#define ULONG_LL			0X00004C4C
#define ULONG_NN			0X00004E4E
#define ULONG_RR			0X00005252
#define ULONG_SG			0X00005347
#define ULONG_SS			0X00005353
#define ULONG_ST			0X00005354
#define ULONG_WW			0X00005757
#define ULONG_DADASH		0X00002D2D      // ASCII equivalent of --
#define ULONG_AAA			0X00414141
#define ULONG_BBB			0X00424242
#define ULONG_OOO			0X004F4F4F
#define ULONG_NOS			0X004E4F53
#define ULONG_QQQ			0X00515151
#define ULONG_WALL		    0X0000003D      // ASCII equivalent of =
#define ULONG_DASH		    0X0000002D      // ASCII equivalent of -
#define ULONG_WALLS		    0X003D3D3D      // ASCII equivalent of ===
#define ULONG_DASHES		0X002D2D2D      // ASCII equivalent of ---
#define ULONG_ZERO  		0X00000000
#define ULONG_BLANK 		0X000000FF
#define ULONG_QUERY 		0X00000063		 // middle byte of DONTKNOW is '?'
#define ULONG_GD 		    0X00004744		 // Sound rewarding 'object'
// The constants enumerated here must correspond to the strings given at
// the beginning of ppworld.cpp. The computer uses the constants, but
// is programmed to display the corresponding strings for the user to see.

// Events are unsigned long constants. Event-types are short constants, given
// by the following enumeration. Every event belongs to at least one event-type.
// Not more than one event of each event-type can occur in each step.
enum event_type_constants 		// Must align with strings in ppworld.cpp <<<!!
	{BODYMOVE, FORWARD, LEFT, RIGHT, TURN,
	 EYEMOVE, EYESTAY, EYENORTH, EYEEAST, EYESOUTH, EYEWEST, NOTYEYEMOVE,
	 BODYEYE,
	 SAY,
	 TOUCH,
	 PROPRIOCEP, FOVEA,
	 WIDE, WIDENONLY, WIDEEONLY, WIDESONLY, WIDEWONLY,
	 DIRECTION,
	 HEAR, HEARSF, HEAROR, HEARFO,
     DESIRE,
//     LANDMARK,
	};

// A template specifies the set of event-types that have to be represented in
// the context and associated event of a valid association.
enum template_constants     // Must align with strings in ppworld.cpp
	{BE,       				// BodyEye action-predicting template
	 SA,					// BodyEye action-predicting template
	 TCH,       			// Feel stimulus-predicting template
	 PR,       				// Proprioception stimulus-predicting template
	 F,                     // Fovea stimulus-predicting templates
	 W,                     // Wide stimulus-predicting templates
	 DN, DNU,                    // Direction of robot B if in fovea or wide
	 SS,                    // Hear stimulus-predicting template
     DE,                    // Desire template
//     LM,                    // LandMark template
	 NOTEMP
	};

// A cluster is a group of templates specifying mutually-exclusive contexts.
// i.e. Only one context of a template in a cluster can match STM in any step.
enum cluster_A_constants    // Action-predicting clusters
  {CLBE, CLSA, NOCLUSTERA
  };

enum cluster_S_constants    // Stimulus-predicting clusters
  {CLTCH, CLPR, CLF,
	 CLW,
	 CLDN, CLDNU,
	 CLSS,
   CLDE,
    NOCLUSTERS
  };

// The event-types are divided up amongst the streams so that no stream can
// have concurrent events. A stream may contain mutually-exclusive event-types.
// i.e. the number of streams is the maximum number of events which can occur
// in the same step.
enum stream_constants
  {BESTR, SASTR,
   TCHSTR, PROPSTR, FOVSTR, WIDESTR, DIRNSTR, SSSTR, DESTR, // LMSTR,
   NOSTREAM
  };

// The event_type struct holds an event and its event-type.
struct event_type  {short type; unsigned long event; };   // 6 types

// STM (short-term memory) holds each current event together with events
// of the same event-type from the past steps as required by the templates.
// This is done in two kinds of shift-register, called "timing" and "threading".
struct event_info   // All information for one event-type in purrpuss_info.
 { char acst,                        // whether action AC or stimulus ST.
		  tim, thr;                    // timing and/or threading shift-register
	unsigned long tim_shift_reg[MAX_EVENT_SHIFTS],  // Current and past events
					  thr_shift_reg[MAX_EVENT_SHIFTS];  // Current and past events
 };                             // 3 + 8 * MAX_EVENT_SHIFTS

// The template priority hierarchy works like this. If the template in the entry
// matches, then proceed to the template pointed to by the *match pointer,
// otherwise proceed to the template pointed to by the *nomatch pointer.
// If a template should always follow another, it will be pointed to by both
// the *match and the *nomatch pointers of the other.
struct temp_priority_entry      // Template pref hierarchy - in stream_struct
 { short temp;                                // TEMPLATE
	struct temp_priority_entry *match,         // MATCH ENTRY
										*nomatch;       // NO MATCH ENTRY
 };                              // 10 bytes

struct nodeA                   // POU. Also referred to as a node
 { float expn;                 // Actual expectation is expn/age.
   long age;                   // node age
   long step_last_used;
   long  pei;
   long  back;                 // first back node slot
   long  ctxtel;               // ctxtA array element pointing to node 4jan01
//	short cluster;	  // Get from pp->templates[temp].clust    removed 23apr99
   short temp;    			   // template     21apr99
   short cntxt_no;             // context number 21apr99
   short next;                 // next context in list     4jan01
 };                              // 30 bytes

struct nodeS                   // POU. Also referred to as a node
 { long age;                      // node age
   long step_last_used;
   long  pei;
   long  ctxtel;                 // ctxtS array element pointing to node 4jan01
//	short cluster;	  // Get from pp->templates[temp].clust    removed 23apr99
   short temp;    					 // template     21apr99
   short cntxt_no;                // context number 21apr99
   short next;                    // next context in list     4jan01
 };                               // 22 bytes

struct predictionA                // PEI. Also referred to as associated event
 { float expn;                    // Actual expectation is expn/age.
	long age;
	short reward_rate_f;   // estimates food reward rate 0-100 stochastically.
	short reward_rate_d;   // estimates drink reward rate 0-100 stochastically.
	char  novelty;			  // 1 novelty, 2 food reward, 4 drink reward	
	char  disapproval;
	short  mark;
	struct event_type event;
	long  ptr_prob;                // number of first ptr_prob.
	long next;
 };                               // 29 bytes

struct predictionS                // PEI. Also referred to as associated event
 { long age;
	struct event_type event;
	long next;
 };                               // 14 bytes

// Used to make linked lists of pointers from a node to nodes which point to it.
struct back_pointer    // See array "bptr[MAX_BACK_POINTERS]" of 'back_pointer's
 { short node;         // preceding node
   long  next;         // next backnode slot.
 };                    // 6 bytes

// Used to make linked lists of pointers to nodes from an associated event.
struct pointer_prob    // See array "pprb[MAX_POINTER_PROBS]" of 'pointer_prob's
 { long prob;
   short  node;                     // next node
   long  next;         // next in list.       size = 10
 };                    // 10 bytes

// Used to build LTM tree of contexts for each template.
struct context_link    // See array "ctxt[]" of 'context_link's
 { struct event_type event;
   long  next,       // next in list
		 next_list,  // next list - context depth
         linkbk;     // back to one pointing to it by next or next_list
 };      			      // 18 bytes                                // size = 12

// Used in choice lists and mark lists to identify nodes with their clusters.
struct node_cluster
 { short node, cluster;
 };                    // 4 bytes

// Used in arrays of event choices with sublists of pred nodes
struct choice_A  // See choice-listA for array of 'choice_A's
 { float worth, age_worth;            // See
	struct event_type event;
	struct node_cluster  chnodesA[MAX_NODES_PER_CHOICE]; // nodes predicting event
 };                    // 14 + 4 * MAX_NODES_PER_CHOICE

struct choice_S  // See choice-listS for array of 'choice_S's
 { float age_worth;
	struct event_type event;
	struct node_cluster  chnodesS[MAX_NODES_PER_CHOICE]; // nodes predicting event
 };                    // 10 + 4 * MAX_NODES_PER_CHOICE

struct rest_choice_S  // For going through alternative stim possibilities
 { short index_best;     // position in choice-list of best for removal
	long rest_worth;
 };                    // 6 bytes

struct choice_listA
 { struct choice_A charray[MAX_CHOICES];
 };                   // (14 + 4 * MAX_NODES_PER_CHOICE) * MAX_CHOICES

struct choice_listS
 { struct choice_S charray[MAX_CHOICES];
 };                     // (10 + 4 * MAX_NODES_PER_CHOICE) * MAX_CHOICES

struct cluster_cntxtsA  // Parts to save in switching between H&N and T&T.
 { short ncntxts;
	short last_node;
	struct event_type  last_event;            // for update_cntxt()
 };                    // 10 bytes

struct cluster_cntxtsS  // Parts to save in switching between H&N and T&T.
 { short ncntxts;
	short last_node;
	struct event_type  last_event;            // for update_cntxt()
 };                    // 10 bytes

struct body_posn
 { short row; short col; unsigned long dirn;
 };                   // 8 bytes

struct board_posn
 { short row; short col;
 };

struct body_info
 {	struct body_posn		 bodypos;
	struct board_posn		 eyepos;
 };                    // 12 bytes

struct m_file_line
 {	long number_step_cycles;
	long A_subacts;
	long A_steps_below_sub;
	long A_reward_fs;
	long A_reward_ds;
	long A_total_fs;
	long A_total_ds;
	long A_reward_gs;
	long B_reward_fs;
	long B_reward_ds;
	long B_total_fs;
	long B_total_ds;
	long B_rewards_nofd;
	long B_steps_below_sub;
    long number_of_B_tells;
	short number_BE_associations;
	short number_f_rewarded_BE_associations;
	short number_d_rewarded_BE_associations;
    short number_fd_rewarded_BE_associations;
	short number_novelty_BE_associations;
//    short landmarks_in_BE_associations;
	short number_plans_followed;
	short number_TCH_associations;
	short number_PR_associations;
	short number_F_associations;
	short number_W_associations;
	short number_DN_associations;
 };                   // 78 bytes

struct reward_getting_association
 {long step;
  short node;
  short temp;
  struct event_type evtyp;
 };                   // 14 bytes

struct hold_info   // Data that must be recovered after alteration in T&T
 { struct cluster_cntxtsA     clustersA[MAX_A_CLUSTERS];
	struct cluster_cntxtsS     clustersS[MAX_S_CLUSTERS];
	struct event_type          current_stimulus[MAX_S_STREAMS],
										current_action[MAX_A_STREAMS];
	struct event_info          events[MAX_EVENT_TYPES];
//c 	unsigned long              canvas[CANVAS_SIZE];
//c 	short							   decay[CANVAS_SIZE];
	float  last_expectation[MAX_A_STREAMS];   // needed for uphill algorithm
	float  current_expectation[MAX_A_STREAMS];// needed for uphill algorithm
	long step;
 };   // 20 * MAX_S_CLUSTERS + 6(MAX_S_STREAMS + MAX_A_STREAMS) + ((3 + 8 * MAX_EVENT_SHIFTS) * MAX_EVENT_TYPES)
	  //   + 8 * MAX_A_STREAMS + 1


// Data associated with clusters.
struct cluster_struct_A               // Cluster of Action-predicting templates
 { char cname[CLUSTER_NAME_LENGTH+1]; // cluster name
	short ntemps;                      // number of templates in cluster
	short highest_node;            // currently largest cntxtn no for cluster
	short last_node;
	short leakback_cluster_no;      // -1 if not a leakback cluster
	short reward_associations;      // nov>1
//    short lm;                       // number of landmarks
	short empty_assA_list;
	long  empty_prdA_list;
	long  empty_pprb_list;
	long  empty_bkptr_list;
	long  empty_ctxtA_list;
//    long  landmark_tAge;   // tAge above which associations become landmarks
//    unsigned long landmark_event;
  short marksA[MAX_MARKS_A]; // Allows all A clusters to be leakback 1jun01
	short nov[MAX_REWARD_NOVELTY+1];  // previously  nov0, nov1 etc
    struct context_link ctxtA[MAX_CONTEXT_LINKS_A];   // Put in cluster if > 1 leakback cluster
	struct nodeA    assA[MAX_NODES_A];
	struct predictionA     prdA[MAX_PREDICTIONS_A];
	struct pointer_prob    pprb[MAX_POINTER_PROBS];
	struct back_pointer    bptr[MAX_BACK_POINTERS];
	struct event_type  last_event;            // for update_cntxt()
	char tem[MAX_TEMPS_IN_CLUSTER];           // template numbers in cluster
 };  // CLUSTER_NAME_LENGTH + 31 + 2*MAX_REWARD_NOVELTY + 18*MAX_CONTEXT_LINKS_A
	 // + 30*MAX_NODES_A + 29*MAX_PREDICTIONS_A + 10*MAX_POINTER_PROBS
	 // + 6*MAX_BACK_POINTERS + 6 + MAX_TEMPS_IN_CLUSTER


struct cluster_struct_S              // Cluster of Stimulus-predicting templates
 { char cname[CLUSTER_NAME_LENGTH+1]; // cluster name
	short ntemps;                      // number of templates in cluster
	short highest_node;              // currently largest cntxtn no for cluster
	short last_node;
	long empty_ctxtS_list;
	long empty_assS_list;
	long empty_prdS_list;
	struct context_link ctxtS[MAX_CONTEXT_LINKS_S];   // Put in cluster if > 1 leakback cluster
	struct nodeS    assS[MAX_NODES_S];
	struct predictionS     prdS[MAX_PREDICTIONS_S];
	struct event_type  last_event;            // for update_cntxt()
	char tem[MAX_TEMPS_IN_CLUSTER];           // template numbers in cluster
 };  // CLUSTER_NAME_LENGTH + 19 + 18*MAX_CONTEXT_LINKS_S
	 // + 22*MAX_NODES_S + 14*MAX_PREDICTIONS_S + 6 + MAX_TEMPS_IN_CLUSTER


struct leakback_cluster   // Extra data for leakback cluster
 { short clusterA;
	short emptyel;             // pointer to first empty location on lkbk_list
	short lkbk_list[LKBK_LISTS_LENGTH];  // Leakback list for cluster.
 };                    // 4 + 2*LKBK_LISTS_LENGTH

struct Leakback_state     // Parameters used in leakback process.
{ short clust,     // Cluster being processed
         node,      // node being processed
         lkbkcl,    // Leakback cluster - numbering different from clusters
         level,     // depth of leakback process
         actvel,    // active element to be processed
         endactel,  // end of active elements on circular list
         emptyel;   // end of empty elements on circular list
};                    // 14 bytes

//struct Novelty_Decrement_Hold
// { short node, cluster;
//	long  pei;
// };

// Data associated with a stream for a robot.
struct stream_struct   // "streams[MAX_STREAMS]" is an array of 'stream_struct's
{ short nevtypes;                                // number of event-types
  short event_type[MAX_EVENTS_IN_STREAM];        // stream event-types
  float last_worth;                      // for incrementing boredom 24oct99
  float last_expectation;                // needed for uphill algorithm
  float current_expectation;             // needed for uphill algorithm.
  struct temp_priority_entry *template_priority;  // for make_choice-lists
};                    // 24 + 6*MAX_EVENTS_IN_STREAM

// Data needed by each template.
struct Template            // "templates[MAX_TEMPLATES]" is array of 'Template's
{ char tname[TEMPLATE_NAME_LENGTH+1];    // template name
  short  clust;                          // cluster to which it belongs
  short  acst;                           // action/stimulus of prediction
  short  npreds;                         // no of prediction types
  short  noEvTypesInContext;             // no of event-types in context
  short  nlinks;                         // no of context links for template
  short  is_in_leakback_cluster;		  // TRUE/FALSE
  char   predict_event_type[MAX_TEMP_PREDICTION_TYPES]; // prediction event-types
  short  eventdeftype[MAX_CONTEXT_EVENTS];    // types in context
  short  eventdefdelay[MAX_CONTEXT_EVENTS];   // event delays in context
  short  eventdef[MAX_CONTEXT_EVENTS];     // ((type << 8) | delay)
  short  kindofpred;                     // choice/recency
  long   base;                           // base of context tree
};  // TEMPLATE_NAME_LENGTH + 19 + MAX_TEMP_PREDICTION_TYPES + 6*MAX_CONTEXT_EVENTS

struct robot_info
{ char    nevtypes,           // largest number of types used
		  nastreams,         // number of action streams
		  nsstreams,         // number of stimulus streams
		  nstreams,				// nastreams + nsstreams
		  mode,              // flag for H_N / T_T, H&N/T&T
		  drive;             // flag for different reward drives 2=food,4=drink
  short   boredom,           //
		  boredom_inc,       //
		  frustration,       //
		  drive_has_changed,
		  reward,            //  2=FOOD for food, 4=DRINK for drink, 8="GOOD"
		  disapproval;       //
  long    step,              // run step
		  steps_with_f_drive_and_O,   // These below don't need to be long <<<<
		  steps_with_d_drive_and_Q,
		  steps_below_sublevel,
		  before_last_body_move,
		  reward_counter_f,    // counts total number of food rewards
		  rewards_without_drivefd, // counts number of rewards with neither drive
		  reward_counter_d,    // counts total number of drink rewards
          reward_counter_gd,   // counts total number of "GOOD" rewards
          number_of_B_tells,   // no times B tells A a move.
		  subsumption_counter; // counts total number of subsumption actions
  float   resource_f,       // starts at 100.0
		  resource_d;       // starts at 100.0
  struct  body_info		 body;      // body and eye positions
  struct  event_type		 touch_evty;
  struct  event_type		 propriocep_evty;
  struct  event_type		 fovea_evty;
  struct  event_type		 wide_evty;
  struct  event_type		 dirn_evty;
  struct  event_type		 hear_evty;
  struct  event_type		 desire_evty;
//  struct  event_type		 landmark_evty;
//c 	struct event_type		 canvas_evty;   // the four corners of canvas[]
  struct  event_info     events[MAX_EVENT_TYPES];	    //
  struct  stream_struct  streams[MAX_STREAMS];         //
  struct  event_type     current_stimulus[MAX_STREAMS],  // Not MAX_S_STREAMS
								 predicted_stimulus[MAX_STREAMS],//     do.
								 current_action[MAX_STREAMS],    // Not MAX_A_STREAMS
								 last_action[MAX_STREAMS],       //     do.
								 predicted_action[MAX_STREAMS];  //     do.
}; // 100 + 6*MAX_EVENT_TYPES + (54 + 6*MAX_EVENTS_IN_STREAM)*MAX_STREAMS

struct pp_info
{ char  ntemplates,        // number of templates                  <<
		naclusters,        // number of action clusters            <<
		nsclusters,        // number of stimulus clusters          <<
		nclusters,         // total number of clusters             <<
		t_t_termination;   // get_actions() and run(). Flag ON if T&T halted. <<
  short lkbk_flag,         // controls return to Windows during leakback
		path_steps_to_go,  // number of steps of plan remaining       <<
		path_step,         // number of step along path.     8mar00
		path_length,       //                                            <<
		marks_seen,        // number of marks seen                   <<
		nlkbckclstrs,      // number of leakback clusters            <<
		nodes_displayed,   // inserts CRLF after every 20. 3sep99    <<
		max_nov,           // max-reward_novelty
		not_learning,      // set to TRUE for no learning or planning
		reward_rate_threshold,
		highest_reward_getting_association,
		last_rewarded_plan_number;
  long  tt_paths_longer_than_2,  //                                    <<
		tt_paths_longer_than_5,  //                                    <<
		paths_longer_than_2_followed,  //                              <<
		lkbck_counter;   // counts total number of leakback calculations <<
  short leak_back_clusters[MAX_CLUSTERS]; // Array of leakback clusters  <<
  short node_register[LKBK_LISTS_LENGTH];
//c   unsigned long         canvas[CANVAS_SIZE];	// for painted vision     <<
//c   short decay[CANVAS_SIZE];						// for painted vision        <<
  short nodecountA[MAX_A_CLUSTERS];
  short maxnodeA[MAX_A_CLUSTERS];
  short nodecountS[MAX_S_CLUSTERS];
  short maxnodeS[MAX_S_CLUSTERS];
  long  predcountA[MAX_A_CLUSTERS];
  long  maxpredA[MAX_A_CLUSTERS];
  long  probcountA[MAX_A_CLUSTERS];
  long  maxprobA[MAX_A_CLUSTERS];
  long  bkptcountA[MAX_A_CLUSTERS];
  long  maxbkptA[MAX_A_CLUSTERS];
  long  predcountS[MAX_S_CLUSTERS];
  long  maxpredS[MAX_S_CLUSTERS];
  long  rewarded_plans[MAX_REWARDED_PLANS];  // steps where rewarded plans made
  struct rest_choice_S rest_choice_list_S[MAX_STREAMS];      // 3mar00
  struct Template       templates[MAX_TEMPLATES];     //                 <<
  struct cluster_struct_A clustersA[MAX_A_CLUSTERS];
  struct cluster_struct_S clustersS[MAX_S_CLUSTERS];
  struct choice_listA   choice_listsA[MAX_A_STREAMS];
  struct choice_A			best_predictionsA[MAX_A_STREAMS];
  struct choice_A       null_choiceA;
  struct choice_listS   choice_listsS[MAX_STREAMS];   // Not MAX_S_STREAMS until
  struct choice_S       best_predictionsS[MAX_STREAMS]; // streams separated
  struct choice_S       null_choiceS;
  struct leakback_cluster leakback_clusters[MAX_LEAKBACK_CLUSTERS];
  struct Leakback_state lkbkst;
//  struct Novelty_Decrement_Hold novdec; // Will need one for each A stream
  struct reward_getting_association rgass[MAX_REWARD_GETTING_ASSOCIATIONS];
  struct hold_info      hold; // to save data that might be altered in T&T  <<
 };   // 49 + 2*MAX_CLUSTERS + 2*LKBK_LISTS_LENGTH + 4*MAX_REWARDED_PLANS
	 // + MAX_A_CLUSTERS*(12 + CLUSTER_NAME_LENGTH + 31 + 2*MAX_REWARD_NOVELTY + 18*MAX_CONTEXT_LINKS_A
	 //  + 30*MAX_NODES_A + 29*MAX_PREDICTIONS_A + 10*MAX_POINTER_PROBS
	 //  + 6*MAX_BACK_POINTERS + 6 + MAX_TEMPS_IN_CLUSTER)
	 // + MAX_S_CLUSTERS*(4 + CLUSTER_NAME_LENGTH + 19 + 18*MAX_CONTEXT_LINKS_S
	 //  + 22*MAX_NODES_S + 14*MAX_PREDICTIONS_S + 6 + MAX_TEMPS_IN_CLUSTER)
	 // + MAX_TEMPLATES*(TEMPLATE_NAME_LENGTH + 19 + MAX_TEMP_PREDICTION_TYPES
	 //  + 6*MAX_CONTEXT_EVENTS)
	 // + MAX_STREAMS*(6 + (10 + 4 * MAX_NODES_PER_CHOICE) * (MAX_CHOICES + 1))
	 // + MAX_A_CLUSTERS*(CLUSTER_NAME_LENGTH + 31 + 2*MAX_REWARD_NOVELTY + 18*MAX_CONTEXT_LINKS_A
	 //  + 30*MAX_NODES_A + 29*MAX_PREDICTIONS_A + 10*MAX_POINTER_PROBS
	 //  + 6*MAX_BACK_POINTERS + 6 + MAX_TEMPS_IN_CLUSTER)
	 // + MAX_S_CLUSTERS*(CLUSTER_NAME_LENGTH + 19 + 18*MAX_CONTEXT_LINKS_S
	 //  + 22*MAX_NODES_S + 14*MAX_PREDICTIONS_S + 6 + MAX_TEMPS_IN_CLUSTER)
	 // + MAX_A_STREAMS*((14 + 4 * MAX_NODES_PER_CHOICE) * (MAX_CHOICES + 1))
	 // + 24 + 8 * MAX_NODES_PER_CHOICE
	 // + 4*MAX_NODES_A + MAX_LEAKBACK_CLUSTERS*(4 + 2*LKBK_LISTS_LENGTH)
	 // + 14 + 14*MAX_REWARD_GETTING_ASSOCIATIONS
	 // + 20 * MAX_S_CLUSTERS + 6(MAX_S_STREAMS + MAX_A_STREAMS) + ((3 + 8 * MAX_EVENT_SHIFTS) * MAX_EVENT_TYPES)
	 //   + 8 * MAX_A_STREAMS + 1


struct world
 { char  robot,              // Active robot: A or B. B is subsumption only
			filechar;           // Char which is doubled at head of file name.
	short	npuss,              // number of PP entities.
			object_squashed,
			quantum,
			food,               //  w.food  = FOOD  in ppcycle.cpp
			drink,              //  w.drink = DRINK in ppcycle.cpp
			plus,               //  w.plus  = PLUS  in ppcycle.cpp
            good,               //  w.good  = GOOD  in ppcycle.cpp 
			reward_count,
			step_cycle_before_counters_zeroed,
			rewards_between_goal_change,
			leakback_from_all_nov_and_reward_nodes,
			leakback_node_limit,
			leakback_level_limit,
			leakback_frequency,        // 1 less than no of steps between leakbacks
			plans_remove_novelty,
			novelty_with_leakback_only,
			expn_calc_omits_ptrs_to_same_node,
			m_file_line,    // Number of current line in Matlab graph array
			last_node_factor,
//            landmark_reciprocal_equilibrium_ratio,
//            landmarks_on,
			robot_B_helps,      // If TRUE Robot B's moves help Robot A
            robot_B_tells,
			diag,					  // flag for printing diagnostics
			jumping,			     // TRUE when jumping to step ahead, FALSE otherwise
			min_file_record_while_jumping, // for fast unrecorded processing
			no_display_or_file, // except step number in new_step(HN)
            holding_reward_getting_associations,
			random_seed,        // if 0, time is used as seed. get_option
			mark_up_factor,     // Used by mark_up()
			stage,              // of main cycle
			deterministic,
			probabilistic;
	unsigned seed;
	long  change_file_step,
			help,					  // no of help actions provided by Robot B.
			number_step_cycles,
			step_cycles_before_showing_all_graph_lines,
			step_for_restart_with_different_file_name,
			jumpStep;			  // step to jump to without Tchr interference
	unsigned long sound_A,      // holds sound originating from robot A
	              sound_B,      // holds sound originating from robot B
                  ul_B_told_A,      // action B told A to do last step
                  ul_B_told_A_prev; // action B told A to do step before last
	float	expectation_for_novelty,
            expectation_for_good,
			discount_factor,
			fooddec,            //  food resource decrement due to forward move
			drinkdec,           //  drink resource decrement per action
			foodinc,            //  food resource increment due to reward
			drinkinc,           //  drink resource increment due to reward
			sat_up,             //  level of resource above which drive vanishes
			sat_down,           //  level of resource below which drive comes on
			sublevel,           //  level of resource below which actions are subsumption
			leakback_threshold;
	 unsigned long      current_goal;
	unsigned long        board[BOARD_CELLS];
	struct board_posn		object, old_object_posn;
	long  m_step_cycles[MAX_FILE_LINES];
	struct m_file_line   m_file[MAX_FILE_LINES];
};  // 152 + 4*BOARD_CELLS + 78*MAX_FILE_LINES

class TWindFile;    // For Windows operation; details in ppfuncs.hpp

#include "c:\Program Files\Borland\CBuilder3\examples\foodrink\ppfuncs.hpp" // needed after above defns.
#endif

